@@ -23,9 +23,6 @@
 #include "config.h"
 
 #include "gkr-daemon.h"
-#include "gkr-daemon-util.h"
-
-#include "common/gkr-location.h"
 
 #include "egg/egg-buffer.h"
 #include "egg/egg-secure-memory.h"
@@ -42,6 +39,9 @@
 #include "ui/gkr-ask-request.h"
 #include "ui/gkr-ask-daemon.h"
 
+#include "util/gkr-daemon-util.h"
+#include "util/gkr-location.h"
+
 #include <unistd.h>
 #include <errno.h>
 #include <stdlib.h>
@@ -549,9 +549,9 @@ request_change_keyring_password (GkrKeyringRequest *req, GkrKeyring* keyring,
 		flags |= GKR_ASK_REQUEST_ORIGINAL_PASSWORD;
 
 	if (is_default) {
-		primary = g_markup_printf_escaped (_("Choose a new password for the '%s' keyring. "), keyring_name);
+		primary = g_markup_printf_escaped (_("Choose a new password for the '%s' keyring."), keyring_name);
 	} else {
-		primary = g_markup_printf_escaped (_("Choose a new password for the default keyring. "));
+		primary = g_markup_printf_escaped (_("Choose a new password for the default keyring."));
 	}
 	
 	/* And put together the ask request */
@@ -1603,6 +1603,8 @@ find_in_each_keyring (GkrKeyring* keyring, gpointer data)
 	GkrKeyringItem *item;
 	GList *ilist;
 	
+	g_object_ref (keyring);
+	
 	for (ilist = keyring->items; ilist != NULL; ilist = ilist->next) {
 		item = ilist->data;
 		if (!gkr_keyring_item_match (item, ctx->type, keyring->locked ? ctx->hashed : ctx->attributes, FALSE))
@@ -1612,7 +1614,7 @@ find_in_each_keyring (GkrKeyring* keyring, gpointer data)
 			
 		if (keyring->locked) {
 			if (!request_keyring_access (ctx->req, keyring))
-				return TRUE;
+				break;
 		}
 
 		if (request_item_access (ctx->req, item, GNOME_KEYRING_ACCESS_READ, TRUE)) {
@@ -1620,7 +1622,9 @@ find_in_each_keyring (GkrKeyring* keyring, gpointer data)
 			ctx->items = g_list_prepend (ctx->items, item);
 		}
 	}
-	
+
+	g_object_unref (keyring);
+
 	return TRUE;
 }
 
