@@ -360,24 +360,34 @@ static void
 lock_key_pair (GP11Session *session, GP11Object *priv, GP11Object *pub)
 {
 	GError *error = NULL;
+	GList *objects, *l;
+
 	g_assert (GP11_IS_SESSION (session));
+	g_assert (GP11_IS_OBJECT (priv));
 	g_assert (GP11_IS_OBJECT (pub));
-	
+
 	if (!login_session (session))
 		return;
 
-	gp11_object_set_session (priv, session);
-	gp11_object_set (priv, &error, CKA_GNOME_AUTH_CACHED, GP11_BOOLEAN, FALSE, GP11_INVALID);
-	if (error != NULL) {
-		g_warning ("couldn't clear cached authentication for key: %s", error->message);
+	/* Delete any authenticator objects */
+	objects = gp11_session_find_objects (session, &error,
+	                                     CKA_CLASS, GP11_ULONG, CKO_GNOME_AUTHENTICATOR,
+	                                     CKA_GNOME_OBJECT, GP11_ULONG, gp11_object_get_handle (priv),
+	                                     GP11_INVALID);
+
+	if (error) {
+		g_warning ("couldn't search for authenticator objects: %s", error->message);
 		g_clear_error (&error);
+		return;
 	}
-	
-	gp11_object_set_session (pub, session);
-	gp11_object_set (pub, &error, CKA_GNOME_AUTH_CACHED, GP11_BOOLEAN, FALSE, GP11_INVALID);
-	if (error != NULL) {
-		g_warning ("couldn't clear cached authentication for key: %s", error->message);
-		g_clear_error (&error);
+
+	/* Delete them all */
+	for (l = objects; l; l = g_list_next (l)) {
+		gp11_object_destroy (l->data, &error);
+		if (error) {
+			g_warning ("couldn't delete authenticator object: %s", error->message);
+			g_clear_error (&error);
+		}
 	}
 }
 
@@ -531,6 +541,59 @@ replace_key_pair (GP11Session *session, GP11Attributes *priv, GP11Attributes *pu
 	return TRUE;
 }
 
+static gboolean
+load_contraints (EggBuffer *buffer, gsize offset, gsize *next_offset,
+                 GP11Attributes *priv, GP11Attributes *pub)
+{
+	GTimeVal tv;
+	guchar constraint;
+	guint32 lifetime;
+	time_t time;
+	struct tm tm;
+	gchar buf[20];
+
+	/*
+	 * Constraints are a byte flag, and optional data depending
+	 * on the constraint.
+	 */
+
+	while (offset < egg_buffer_length (buffer)) {
+		if (!egg_buffer_get_byte (buffer, offset, &offset, &constraint))
+			return FALSE;
+
+		switch (constraint) {
+		case GCK_SSH_FLAG_CONSTRAIN_LIFETIME:
+			if (!egg_buffer_get_uint32 (buffer, offset, &offset, &lifetime))
+				return FALSE;
+
+			g_get_current_time (&tv);
+			time = tv.tv_sec + lifetime;
+
+			if (!gmtime_r (&time, &tm))
+				g_return_val_if_reached (FALSE);
+
+			if (!strftime(buf, sizeof (buf), "%Y%m%d%H%M%S00", &tm))
+				g_return_val_if_reached (FALSE);
+
+			gp11_attributes_add_data (pub, CKA_GNOME_AUTO_DESTRUCT, buf, 16);
+			gp11_attributes_add_data (priv, CKA_GNOME_AUTO_DESTRUCT, buf, 16);
+			break;
+
+		case GCK_SSH_FLAG_CONSTRAIN_CONFIRM:
+			/* We can't use prompting as access control on an insecure X desktop */
+			g_message ("prompt constraints are not supported.");
+			return FALSE;
+
+		default:
+			g_message ("unsupported constraint or other unsupported data");
+			return FALSE;
+		}
+	}
+
+	*next_offset = offset;
+	return TRUE;
+}
+
 /* -----------------------------------------------------------------------------
  * OPERATIONS
  */
@@ -580,9 +643,6 @@ op_add_identity (GckSshAgentCall *call)
 		return FALSE;
 	}
 		
-		
-	/* TODO: Blinding? See ssh-agent.c */
-
 	/* Get the comment */
 	if (!egg_buffer_get_string (call->req, offset, &offset, &comment, (EggBufferAllocator)g_realloc)) {
 		gp11_attributes_unref (pub);
@@ -593,7 +653,14 @@ op_add_identity (GckSshAgentCall *call)
 	gp11_attributes_add_string (pub, CKA_LABEL, comment);
 	gp11_attributes_add_string (priv, CKA_LABEL, comment);
 	g_free (comment);
-		
+
+	/* Any constraints on loading the key */
+	if (!load_contraints (call->req, offset, &offset, priv, pub)) {
+		gp11_attributes_unref (pub);
+		gp11_attributes_unref (priv);
+		return FALSE;
+	}
+
 	/* 
 	 * This is the session that owns these objects. Only 
 	 * one thread can use it at a time. 
@@ -638,6 +705,13 @@ op_v1_add_identity (GckSshAgentCall *call)
 	gp11_attributes_add_string (priv, CKA_LABEL, V1_LABEL);
 	gp11_attributes_add_string (pub, CKA_LABEL, V1_LABEL);
 	
+	/* Any constraints on loading the key */
+	if (!load_contraints (call->req, offset, &offset, priv, pub)) {
+		gp11_attributes_unref (pub);
+		gp11_attributes_unref (priv);
+		return FALSE;
+	}
+
 	/* 
 	 * This is the session that owns these objects. Only 
 	 * one thread can use it at a time. 
@@ -1228,6 +1302,6 @@ const GckSshAgentOperation gck_ssh_agent_operations[GCK_SSH_OP_MAX] = {
      op_not_implemented_success,                 /* GKR_SSH_OP_LOCK */
      op_not_implemented_success,                 /* GKR_SSH_OP_UNLOCK */
      op_v1_add_identity,                         /* GKR_SSH_OP_ADD_RSA_ID_CONSTRAINED */
-     op_not_implemented_failure,                 /* GKR_SSH_OP_ADD_ID_CONSTRAINED */
+     op_add_identity,                            /* GKR_SSH_OP_ADD_ID_CONSTRAINED */
      op_not_implemented_failure,                 /* GKR_SSH_OP_ADD_SMARTCARD_KEY_CONSTRAINED */
 };
